/*
 * @author Richard Gradischnegg
 * Created on Oct 31, 2003
 */
package isac.util.parser;

import isac.util.CalcChanged;
import isac.util.Message;
import isac.util.ResponseWrapper;
import isac.util.formulae.Assumptions;
import isac.util.formulae.CalcHead;
import isac.util.formulae.KEStoreID;
import isac.util.formulae.Model;
import isac.util.formulae.ModelItem;
import isac.util.formulae.ModelItemList;
import isac.util.formulae.CalcHeadSimpleID;
import isac.util.formulae.FormHeadsContainer;
import isac.util.formulae.Formula;
import isac.util.formulae.CalcFormula;
import isac.util.formulae.Position;
import isac.util.formulae.Specification;
import isac.util.tactics.Rewrite;
import isac.util.tactics.RewriteInst;
import isac.util.tactics.RewriteSet;
import isac.util.tactics.RewriteSetInst;
import isac.util.tactics.SimpleTactic;
import isac.util.tactics.StringListTactic;
import isac.util.tactics.SubProblemTactic;
import isac.util.tactics.TacticsContainer;
import isac.util.tactics.Theorem;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.StringReader;

import org.apache.commons.digester.Digester;
import org.apache.log4j.Logger;
import org.xml.sax.EntityResolver;
import org.xml.sax.ErrorHandler;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

/**
 * parses Objects from XML-Output of the Kernel.
 * 
 * @author Richard Gradischnegg
 */
public class XMLParserDigest {

    private Digester digester_;

    private String dtd_path_;

    class BridgeErrorHandler implements ErrorHandler {

        public void error(SAXParseException e) {
            System.out.println("---xml-error " + e);
        }

        public void fatalError(SAXParseException e) {
            System.out.println("---xml-fatalerror " + e);
        }

        public void warning(SAXParseException e) {
            System.out.println("---xml-warning " + e);
        }
    }

    // auxillary class for feeding the dtd_
    class BridgeEntityResolver implements EntityResolver {

        String dtd_;

        public BridgeEntityResolver(String dtd) {
            this.dtd_ = dtd;
        }

        public InputSource resolveEntity(String publicId, String systemId)
                throws SAXException, IOException {
            if ((systemId != null) && systemId.endsWith("isac.dtd")) {
                return new InputSource(new StringReader(dtd_));
            } else {
                System.out.println("BridgeEntityResolver: return null");
                return null;
            }
        }
    }

    static Logger logger_ = Logger.getLogger(XMLParserDigest.class.getName());

    /**
     * Constructor: create a new XMLParserDigest object and fill it with the
     * rules.
     * 
     * @param dtd_path_
     *            Path to the DTD-File (Document Type Description)
     */
    public XMLParserDigest(String dtdPath) {
        logger_.debug("XMLParserDigest: dtd_path_=" + dtdPath);
        digester_ = new Digester();
        digester_.setErrorHandler(new BridgeErrorHandler());
        this.dtd_path_ = dtdPath;
        //dtd_-File --> String (faster validating)
        StringBuffer sb;
        try {
            BufferedReader br = new BufferedReader(new FileReader(dtdPath));
            sb = new StringBuffer();
            while (br.ready()) {
                sb.append(br.readLine() + "\n");
            }
            // Set dtd_ for parser validation
            digester_
                    .setEntityResolver(new BridgeEntityResolver(sb.toString()));
        } catch (FileNotFoundException e) {
            e.printStackTrace();
        } catch (IOException e) {
            e.printStackTrace();
        }

        // Rules for the digester_ to create Objects and fill them from
        // XML-Files
        digester_.addObjectCreate("ISAC", ResponseWrapper.class);
        digester_.addObjectCreate("*/CALCID", CalcHeadSimpleID.class);
        digester_.addCallMethod("*/CALCID", "setID", 0);
        digester_.addSetRoot("*/CALCID", "setCalcID");

        // Rules for parsing Position
        digester_.addObjectCreate("*/POSITION", Position.class);
        digester_.addCallMethod("*/POSITION/INTLIST/INT", "addInt", 0);
        digester_.addCallMethod("*/POSITION/POS", "setKind", 0);
        digester_.addSetNext("*/POSITION", "setResponse");
        //		digester_.addSetNext("*/POSITION", "setPosition");

        // 3 different Positions in CalcChangedEvent
        digester_.addObjectCreate("*/UNCHANGED", Position.class);
        digester_.addCallMethod("*/UNCHANGED/INTLIST/INT", "addInt", 0);
        digester_.addCallMethod("*/UNCHANGED/POS", "setKind", 0);
        digester_.addSetNext("*/UNCHANGED", "setLastUnchanged");

        digester_.addObjectCreate("*/DELETED", Position.class);
        digester_.addCallMethod("*/DELETED/INTLIST/INT", "addInt", 0);
        digester_.addCallMethod("*/DELETED/POS", "setKind", 0);
        digester_.addSetNext("*/DELETED", "setLastDeleted");

        digester_.addObjectCreate("*/GENERATED", Position.class);
        digester_.addCallMethod("*/GENERATED/INTLIST/INT", "addInt", 0);
        digester_.addCallMethod("*/GENERATED/POS", "setKind", 0);
        digester_.addSetNext("*/GENERATED", "setLastGenerated");

        // Rules for parsing Formula
        digester_.addObjectCreate("*/FORMULA", Formula.class);
        //CalcHeadItems have no <FORMULA> ...
        digester_.addCallMethod("*/MATHML/ISA", "setText", 0);
        digester_.addSetNext("*/FORMULA", "setFormula");

        // Rules for parsing CalcFormula with Position
        digester_.addObjectCreate("*/CALCFORMULA", CalcFormula.class);
        digester_.addSetNext("*/CALCFORMULA", "setResponse");

        //============== Rules for parsing CalcHead
        digester_.addObjectCreate("*/CALCHEAD", CalcHead.class);
        digester_.addSetProperties("*/CALCHEAD", "status", "statusString");

        digester_.addObjectCreate("*/HEAD/MATHML", Formula.class);
        digester_.addSetNext("*/HEAD/MATHML", "setHeadLine");

        digester_.addCallMethod("*/CALCHEAD/BELONGSTO", "setBelongsTo", 0);

        digester_.addObjectCreate("*/CALCHEAD/MODEL", Model.class);
        //..........................................................WN050624xx
        addItemList(digester_, "Given");
        addItemList(digester_, "Where");
        addItemList(digester_, "Find");
        addItemList(digester_, "Relate");
        digester_.addSetNext("*/CALCHEAD/MODEL", "setModel");
        //..........................................................WN050624xx

        digester_.addObjectCreate("*/CALCHEAD/SPECIFICATION",
                Specification.class);
        digester_.addObjectCreate("*/CALCHEAD/SPECIFICATION/THEORYID",
                CalcHeadSimpleID.class);
        digester_
                .addCallMethod("*/CALCHEAD/SPECIFICATION/THEORYID", "setID", 0);
        digester_.addSetNext("*/CALCHEAD/SPECIFICATION/THEORYID", "setTheory");
        digester_.addObjectCreate("*/CALCHEAD/SPECIFICATION/PROBLEMID",
                KEStoreID.class);
        digester_.addCallMethod(
                "*/CALCHEAD/SPECIFICATION/PROBLEMID/STRINGLIST/STRING",
                "addString", 0);
        digester_
                .addSetNext("*/CALCHEAD/SPECIFICATION/PROBLEMID", "setProblem");
        digester_.addObjectCreate("*/CALCHEAD/SPECIFICATION/METHODID",
                KEStoreID.class);
        digester_.addCallMethod(
                "*/CALCHEAD/SPECIFICATION/METHODID/STRINGLIST/STRING",
                "addString", 0);
        digester_.addSetNext("*/CALCHEAD/SPECIFICATION/METHODID", "setMethod");
        digester_.addSetNext("*/CALCHEAD/SPECIFICATION", "setSpecification");
        digester_.addSetNext("*/CALCHEAD", "setResponse");

        //============== Rules for parsing various Tactic Objects
        //	Rules for parsing SimpleTactic
        digester_.addObjectCreate("*/SIMPLETACTIC", SimpleTactic.class);
        digester_.addCallMethod("*/SIMPLETACTIC/MATHML/ISA", "setArgument", 0);
        digester_.addSetProperties("*/SIMPLETACTIC", "name", "name");
        digester_.addSetNext("*/SIMPLETACTIC", "setResponse");

        //	Rules for parsing KeyListTactic
        digester_.addObjectCreate("*/STRINGLISTTACTIC", StringListTactic.class);
        digester_.addCallMethod("*/STRINGLISTTACTIC/STRINGLIST/STRING",
                "addKey", 0);
        digester_.addSetProperties("*/STRINGLISTTACTIC", "name", "name");
        digester_.addSetNext("*/STRINGLISTTACTIC", "setResponse");

        //  Rules for parsing Theorem
        digester_.addObjectCreate("*/THEOREM", Theorem.class);
        digester_.addCallMethod("*/THEOREM/ID", "setId", 0);
        digester_
                .addCallMethod("*/THEOREM/FORMULA/MATHML/ISA", "setFormula", 0);
        digester_.addSetNext("*/THEOREM", "setTheorem");

        //  Rules for parsing RewriteTactic
        digester_.addObjectCreate("*/REWRITETACTIC", Rewrite.class);
        digester_.addSetProperties("*/REWRITETACTIC", "name", "name");
        digester_.addSetNext("*/REWRITETACTIC", "setResponse");

        //  Rules for parsing RewriteInstTactic
        digester_.addObjectCreate("*/REWRITEINSTTACTIC", RewriteInst.class);
        digester_.addSetProperties("*/REWRITEINSTTACTIC", "name", "name");
        digester_.addSetNext("*/REWRITEINSTTACTIC", "setResponse");
        digester_.addCallMethod(
                "*/REWRITEINSTTACTIC/SUBSTITUTION/PAIR/VARIABLE/MATHML/ISA",
                "setVariable", //WN0501 ^^^ this seems to work only on
                // spec.case
                0);
        digester_.addCallMethod(
                "*/REWRITEINSTTACTIC/SUBSTITUTION/PAIR/VALUE/MATHML/ISA",
                "setValue", //WN0501 ^^^ this seems to work only on spec_.case
                0);

        //	Rules for parsing RewriteSetTactic
        digester_.addObjectCreate("*/REWRITESETTACTIC", RewriteSet.class);
        digester_.addCallMethod("*/REWRITESETTACTIC/RULESET", "setRuleSet", 0);
        digester_.addSetProperties("*/REWRITESETTACTIC", "name", "name");
        digester_.addSetNext("*/REWRITESETTACTIC", "setResponse");

        //	Rules for parsing RewriteSetInstTactic
        digester_.addObjectCreate("*/REWRITESETINSTTACTIC",
                RewriteSetInst.class);
        digester_.addCallMethod("*/REWRITESETINSTTACTIC/RULESET", "setRuleSet",
                0);
        digester_.addCallMethod(
                "*/REWRITESETINSTTACTIC/SUBSTITUTION/PAIR/VARIABLE/MATHML/ISA",
                "setVariable", //WN0501 ^^^ this seems to work only on
                // spec_.case
                0);
        digester_.addCallMethod(
                "*/REWRITESETINSTTACTIC/SUBSTITUTION/PAIR/VALUE/MATHML/ISA",
                "setValue", //WN0501 ^^^ this seems to work only on spec_.case
                0);
        digester_.addSetProperties("*/REWRITESETINSTTACTIC", "name", "name");
        digester_.addSetNext("*/REWRITESETINSTTACTIC", "setResponse");

        // Rules for parsing SubProblemTactic
        digester_.addObjectCreate("*/SUBPROBLEMTACTIC", SubProblemTactic.class);
        digester_.addCallMethod("*/SUBPROBLEMTACTIC/THEORY", "setTheoryID", 0);
        digester_.addCallMethod("*/SUBPROBLEMTACTIC/PROBLEM/STRINGLIST/STRING",
                "addProblemString", 0);
        digester_.addSetProperties("*/SUBPROBLEMTACTIC", "name", "name");
        digester_.addSetNext("*/SUBPROBLEMTACTIC", "setResponse");

        //	Rules for parsing CalcIterator move message
        digester_.addObjectCreate("*/CALCITERATOR/ERROR", Message.class);
        digester_.addCallMethod("*/CALCITERATOR/ERROR", "setError", 0);
        digester_.addSetNext("*/CALCITERATOR/ERROR", "setResponse");

        //	Rules for parsing CalcIterator constructor message
        digester_.addObjectCreate("*/ADDUSER", CalcHeadSimpleID.class);
        digester_.addCallMethod("*/ADDUSER/USERID", "setID", 0);
        digester_.addSetNext("*/ADDUSER", "setResponse");

        //  Rules for parsing autoCalculateResponse
        digester_.addCallMethod("*/AUTOCALC/ERROR", "setError", 0);
        digester_.addObjectCreate("*/CALCCHANGED", CalcChanged.class);
        digester_.addSetNext("*/CALCCHANGED", "setResponse");

        // Rules for parsing fetchApplicableTactics
        digester_.addObjectCreate("*/APPLICABLETACTICS/TACLIST",
                TacticsContainer.class);
        digester_.addSetNext("*/APPLICABLETACTICS/TACLIST", "setResponse");

        // Rules for getElementsFromTo
        digester_.addObjectCreate("*/GETELEMENTSFROMTO/FORMHEADS",
                FormHeadsContainer.class);
        digester_.addSetNext("*/GETELEMENTSFROMTO/FORMHEADS", "setResponse");

        // getAssumptions (and still getAccumulatedAsms)
        digester_.addObjectCreate("*/ASMLIST", Assumptions.class);
        digester_.addSetNext("*/ASMLIST", "setResponse");
    }

    /**
     * parse a ModelItemList
     * 
     * @param modPart
     *            is "GIVEN" or "WHERE" or "FIND" or "RELATE"
     */
    private void addItemList(Digester digester, String modPart) {
        String modPartUp = modPart.toUpperCase();
        // Rules for parsing ItemLists in the model part of a CalcHead: Given
        // Find Where Relate
        digester.addObjectCreate("*/CALCHEAD/MODEL/" + modPartUp,
                ModelItemList.class);
        digester.addObjectCreate("*/CALCHEAD/MODEL/" + modPartUp + "/ITEM",
                ModelItem.class);
        digester.addSetProperties("*/CALCHEAD/MODEL/" + modPartUp + "/ITEM",
                "status", "itemStatus");
        digester.addCallMethod("*/CALCHEAD/MODEL/" + modPartUp
                + "/ITEM/MATHML/ISA", "setText", 0);
        digester.addSetNext("*/CALCHEAD/MODEL/" + modPartUp + "/ITEM",
                "addItem");
        digester.addSetNext("*/CALCHEAD/MODEL/" + modPartUp, "set" + modPart);
    }

    /**
     * Set whether or not using the DTD for validating
     * 
     * @param validating
     *            <UL>
     *            <LI>true: XML is tested for validity against DTD
     *            <LI>
     *            <LI>false: XML is only tested for well-formed-ness
     *            <LI>
     *            </UL>
     */
    public void setValidating(boolean validating) {
        logger_.warn("setValidating: validating=" + validating);
        digester_.setValidating(validating);
    }

    /**
     * Parse an XML string to Java Object(s)
     * 
     * @param xmlString
     *            the string containing the XML content
     * @return ResponseWrapper containing the CalcID and the parsed Object
     */
    public ResponseWrapper parse(String xmlString) {
        Object o = null;
        xmlString = "<?xml version=\"1.0\"?><!DOCTYPE ISAC SYSTEM \""
                + dtd_path_ + "\"><ISAC>" + xmlString + "</ISAC>";

        encodeEntities(xmlString);
        try {
            digester_.clear();
            o = digester_.parse(new InputSource(new StringReader(xmlString)));
        } catch (IOException e) {
            e.printStackTrace();
        } catch (SAXException e) {
            e.printStackTrace();
        }
        ResponseWrapper wrapper = (ResponseWrapper) o;
        //System.out.println(wrapper.getCalcID().toSMLString());
        return wrapper;
    }

    public static void encodeEntities(String xmlString) {
        xmlString = xmlString.replaceAll(" & ", "&amp;");
        xmlString = xmlString.replaceAll(" > ", "&gt;");
        xmlString = xmlString.replaceAll(" < ", "&lt;");
    }

    public static void decodeEntities(String xmlString) {
        xmlString = xmlString.replaceAll("&amp;", " & ");
        xmlString = xmlString.replaceAll("&gt;", " > ");
        xmlString = xmlString.replaceAll("&lt;", " < ");
    }
}