package isac.wsdialog;

import isac.bridge.*;
import isac.interfaces.*;
import isac.session.ISessionManager;
import isac.useractions.IUIAction;
import isac.useractions.IUserAction;
import isac.useractions.UIAction;
import isac.useractions.UIActionOnCalcElement;
import isac.useractions.UIActionOnIterator;
import isac.useractions.UserAction;
import isac.useractions.UserActionOnCalcElement;
import isac.useractions.UserActionOnInt;
import isac.useractions.UserActionOnIterator;
import isac.util.*;
import isac.util.formulae.*;
import isac.util.tactics.*;
import isac.util.users.IUserSettings;
import isac.util.users.User;
import isac.util.users.UserSettings;

import java.io.Serializable;
import java.net.MalformedURLException;
import java.rmi.*;
import java.rmi.registry.LocateRegistry;
import java.rmi.server.UnicastRemoteObject;
import java.util.Vector;

import org.apache.log4j.Logger;

/**
 * The WorksheetDialog moderates the communication between two instances working
 * on the same CalcTree object. One of these instances is the user, the other a
 * math engine. Most probably, the user's GUI, the WorksheetDialog and the math
 * engine reside on different machines and communicate with each other by means
 * of RMI.
 * 
 * @author Alan Krempler
 * @author Manuel Koschuch
 */
public class WorksheetDialog extends UnicastRemoteObject implements
        IWorksheetDialog, IToUser, Serializable,
        IBrowserDialogToWorksheetDialog {

    static Logger logger_ = Logger.getLogger(BridgeRMI.class.getName());

    protected MathEngine math_engine_;

    protected CalcTree calc_tree_;

    protected ICalcIterator active_formula_;

    /**
     * determines if the caclulation on the worksheet is ready
     */
    protected boolean worksheet_ready_;

    // protected CalcHead calc_head_sub_;

    protected int phase_;

    private Vector datachange_listeners_;

    private IToGUI ui_control_listener_;

    protected String worksheet_id_;

    protected UserSettings user_settings_;// hold whole hashmap locally

    /**
     * @param ME_path
     *            URL of the math engine to be used
     * @param user_path
     *            path(without URL) to UserSettings, UserModel
     * @param session_dialog
     * @param session_id
     *            which create the WorksheetDialog and Worksheet
     * @throws RemoteException
     */
    public WorksheetDialog(String ME_path, ISessionManager session_dialog,
            String session_id, String worksheet_id) throws RemoteException {
        super();
        this.rmiBind();
        datachange_listeners_ = new Vector();
        MathEngine.init(ME_path);
        math_engine_ = MathEngine.getMathEngine();
        phase_ = DIALOGPHASE_IDLE;

        User user = null;
        try {
            user = session_dialog.getUser(session_id);
        } catch (RemoteException e1) {
            e1.printStackTrace();
        }
        user_settings_ = new UserSettings(user.getUserpath(), user
                .getUsername());
        worksheet_id_ = worksheet_id;
    }

    public String getWorksheetID(){
        return worksheet_id_;
    }
    
    
    private void buildUIElementsContextCalcElements() {
        try {
            ui_control_listener_.addUIElement(new UIAction(
                    UI_SOLVE_GET_APPLICABLE_TACTICS,
                    IUIAction.UI_CONTEXT_CALCELEMENTS));
            ui_control_listener_.addUIElement(new UIAction(
                    UI_SOLVE_SHOW_ASSUMPTIONS,
                    IUIAction.UI_CONTEXT_CALCELEMENTS));
            ui_control_listener_
                    .addUIElement(new UIAction(UI_SOLVE_TACTIC_APPLIED,
                            IUIAction.UI_CONTEXT_CALCELEMENTS));
            ui_control_listener_.addUIElement(new UIAction(
                    UI_SOLVE_SHOW_INTERMEDIATE_STEPS,
                    IUIAction.UI_CONTEXT_CALCELEMENTS));
        } catch (RemoteException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }
    }

    /**
     * start a calculation
     */
    // user id hier notwendig? vgl. constructor
    public CalcTree getCalcTree(int user_id, Formalization f, int start_from/*
                                                                             * WN050421 ,
                                                                             * int
                                                                             * requested_calchead_view
                                                                             */) {
        // do administrative stuff
        calc_tree_ = math_engine_.getCalcTree(f);
        calc_tree_.addDataChangeListener(this);

        // have the user or the engine do the specifying

        buildUIElementsContextCalcElements();
        // 050727 --> DialogIterator
        switch (user_settings_.getValue(IUserSettings.KEY1_START_SPECIFY)) {
        case IUserSettings.VAL1_SKIP_SPECIFY_TO_START_SOLVE:
            startSolving();
            break;
        case IUserSettings.VAL1_POP_CALCHEAD_WITH_DESCRIPTIONS:
            phase_ = DIALOGPHASE_SPECIFY;
        }

        return calc_tree_;
    }

    /**
     * at the beginning of a calculation as well as of a subproblem;
     * 
     * @return CalcHead eg. for providing CalcHead#VIEWSTYLE_*
     * 
     * @see isac.wsdialog.IWorksheetDialog#startSpecifying() FIXXXME.WN050704
     *      should be private !!!
     */
    private void startSpecifying(CalcHead calc_head) {
        phase_ = DIALOGPHASE_SPECIFY;
        try {
            // remove solving buttons

            removeNextAndAutoButtonForWorksheet();

            // WN050728
            // instead calc_tree_.autoCalculate(IToCalc.SCOPE_CALCULATION,
            // 1);
            // ...NEW Model_Problem (inserting Descriptions into the Model)
            // without causing CalcChangedEvent (would loop in calcChanged!)
            doUIAction(new UIActionOnCalcElement(UI_SPECIFY_OPEN_CALCHEAD,
                    IUIAction.UI_CONTEXT_SWITCH_PHASES, calc_head));

            if (calc_head.getStatusString().equals("incorrect")) {

                addCalcHeadPanelButtons();
                addNextAndAutoButtonsForCalchead();

            } else {
                // Button Close
                ui_control_listener_.addUIElement(new UIAction(
                        UI_SPECIFY_CLOSE_CALCHEAD,
                        IUIAction.UI_CONTEXT_CALCHEAD));
                addNextAndAutoButtonForWorksheet();
            }

        } catch (RemoteException e) {
            e.printStackTrace();
        }

    }

    private void addCalcHeadPanelButtons() throws RemoteException {
        // Button Complete
        ui_control_listener_.addUIElement(new UIAction(
                UI_SPECIFY_COMPLETE_CALCHEAD, IUIAction.UI_CONTEXT_CALCHEAD));
        // Button Check
        ui_control_listener_.addUIElement(new UIAction(
                UI_SPECIFY_CHECK_CALCHEAD, IUIAction.UI_CONTEXT_CALCHEAD));
        // Button reset
        ui_control_listener_.addUIElement(new UIAction(UI_SPECIFY_RESET,
                IUIAction.UI_CONTEXT_CALCHEAD));
        // Button Start Solving
        ui_control_listener_.addUIElement(new UIAction(UI_SPECIFY_TO_SOLVE,
                IUIAction.UI_CONTEXT_CALCHEAD));

    }

    private void removeNextAndAutoButtonForWorksheet() throws RemoteException {
        // Button next
        ui_control_listener_.removeUIElement(new UIAction(UI_SOLVE_CALCULATE_1,
                IUIAction.UI_CONTEXT_CALCULATION));
        // Button auto
        ui_control_listener_.removeUIElement(new UIAction(
                UI_SOLVE_CALCULATE_ALL, IUIAction.UI_CONTEXT_CALCULATION));

    }

    /**
     * add next and auto Button to WindowApplication
     */
    // FIXME MK Context Calculation is used instead of context calchead,
    // because buttons have to be added to windowapplication and not to
    // calcheadpanel
    private void addNextAndAutoButtonsForCalchead() {
        try {
            // add next Button
            ui_control_listener_.addUIElement(new UIAction(
                    UI_SPECIFY_COMPLETE_STEPWISE,
                    IUIAction.UI_CONTEXT_CALCULATION));
            // add auto Button
            ui_control_listener_
                    .addUIElement(new UIAction(UI_SPECIFY_CALCULATE_ALL,
                            IUIAction.UI_CONTEXT_CALCULATION));
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

    // AK050711 separate end Phase X | start Phase Y
    // endSpecifying()
    // {
    // remove UI Elements
    // dialog_phase_=DIALOGPHASE_IDLE;
    // }

    // /**
    // * @deprecated Use {@link #notifyUserAction(IUserAction)}instead
    // */
    // public void modifyCalcHead(CalcHead calc_head) {
    // // MathEngine.getMathEngine().modifyCalcHead(calc_head);
    // calc_tree_.modifyCalcHead(calc_head);
    // }

    /**
     * @see isac.wsdialog.IWorksheetDialog#startCalculate()
     */
    public void startCalculate() {
        calc_tree_.startCalculate();
    }

    /**
     * dismiss the CalcHeadPanel and start WorksheetDialog#DIALOGPHASE_SOLVE
     * 
     * @see isac.interfaces.IToCalc#startSolving()
     */
    private void startSolving() {
        try {
            // remove specify buttons
            // remove CalcHead Panel
            // doUIAction(new
            // UIAction(UI_SPECIFY_SHOW_CALCHEAD,IUIAction.UI_CONTEXT_SWITCH_PHASES));
            // remove buttons

            removeCalcHeadPanelButtons();
            removeNextAndAutoButtonsForCalchead();
            addNextAndAutoButtonForWorksheet();

        } catch (RemoteException e) {
            e.printStackTrace();
        }
        phase_ = DIALOGPHASE_SOLVE;
        // try {
        // calc_tree_.startSolving();
        // } catch (AutoCalculateException e1) {
        // e1.printStackTrace();
        // }

    }

    private void addNextAndAutoButtonForWorksheet() throws RemoteException {
        // Button next
        ui_control_listener_.addUIElement(new UIAction(UI_SOLVE_CALCULATE_1,
                IUIAction.UI_CONTEXT_CALCULATION));
        // Button auto
        ui_control_listener_.addUIElement(new UIAction(UI_SOLVE_CALCULATE_ALL,
                IUIAction.UI_CONTEXT_CALCULATION));
    }

    private void removeCalcHeadPanelButtons() throws RemoteException {
        // Button Complete
        ui_control_listener_.removeUIElement(new UIAction(
                UI_SPECIFY_COMPLETE_CALCHEAD, IUIAction.UI_CONTEXT_CALCHEAD));
        // Button Start Solving
        ui_control_listener_.removeUIElement(new UIAction(UI_SPECIFY_TO_SOLVE,
                IUIAction.UI_CONTEXT_CALCHEAD));
        // Button Check
        ui_control_listener_.removeUIElement(new UIAction(
                UI_SPECIFY_CHECK_CALCHEAD, IUIAction.UI_CONTEXT_CALCHEAD));
        // Button reset
        ui_control_listener_.removeUIElement(new UIAction(UI_SPECIFY_RESET,
                IUIAction.UI_CONTEXT_CALCHEAD));
    }

    /**
     *  
     */
    private void removeNextAndAutoButtonsForCalchead() {

        // FIXME MK Context Calculation is used instead of context calchead,
        // because buttons have to be added to windowapplication and not to
        // calcheadpanel
        try {
            // add next Button
            ui_control_listener_.removeUIElement(new UIAction(
                    UI_SPECIFY_COMPLETE_STEPWISE,
                    IUIAction.UI_CONTEXT_CALCULATION));
            // add auto Button
            ui_control_listener_
                    .removeUIElement(new UIAction(UI_SPECIFY_CALCULATE_ALL,
                            IUIAction.UI_CONTEXT_CALCULATION));
        } catch (RemoteException e) {
            e.printStackTrace();
        }
    }

    /**
     * accept the changes in a calculation reported by the bridge
     * 
     * ... and pass them to the Worksheet
     * 
     * @see isac.wsdialog.IToUser#calcChanged(isac.wsdialog.CalcChangedEvent)
     */
    public void calcChanged(CalcChangedEvent event) throws RemoteException {
        worksheet_ready_ = false;
        logger_.fatal(" WS<-. . : calcChanged (unc="
                + event.getLastUnchangedFormula().toSMLString() + ", del="
                + event.getLastDeletedFormula().toSMLString() + ", gen="
                + event.getLastGeneratedFormula().toSMLString());
        try {
            for (int i = 0; i < datachange_listeners_.size(); i++) {
                CalcChangedEvent wrappedEvent = new CalcChangedEvent(event
                        .getTransactionId(), event.isCompleted(),
                        new DialogIterator((CalcIterator) event
                                .getLastUnchangedFormula(), this),
                        new DialogIterator((CalcIterator) event
                                .getLastDeletedFormula(), this),
                        new DialogIterator((CalcIterator) event
                                .getLastGeneratedFormula(), this));
                ((IToUser) datachange_listeners_.elementAt(i))
                        .calcChanged(wrappedEvent);
            }
        } catch (Exception e) {
            e.printStackTrace();
        }
        // different call in different phases???
        addUIElementsContextOneElement(event);
        active_formula_ = event.getLastGeneratedFormula();
        if (active_formula_.onCalcHead()) {
            switch (user_settings_.getValue("KEY1_START_SPECIFY")) {
            case IUserSettings.VAL1_SKIP_SPECIFY_TO_START_SOLVE:
                // WN050728
                // instead calc_tree_.autoCalculate(IToCalc.SCOPE_CALCULATION,
                // 1);
                // ...NEW Model_Problem (inserting Descriptions into the Model)
                // without causing CalcChangedEvent (would loop in calcChanged!)
                startSolving();
                break;
            case IUserSettings.VAL1_POP_CALCHEAD_WITH_DESCRIPTIONS:
                startSpecifying((CalcHead) active_formula_.getElement());
                break;
            default:
                try {
                    throw new Exception("WorksheetDialog#calcChanged:switch"
                            + " (user_settings_.getValue(KEY1_START_SPECIFY)");
                } catch (Exception e1) {
                    e1.printStackTrace();
                }
            }
        }
    }

    private void addUIElementsContextOneElement(CalcChangedEvent event) {
        ICalcIterator lastUnchanged = event.getLastUnchangedFormula();
        ICalcIterator lastDeleted = event.getLastDeletedFormula();
        ICalcIterator lastGenerated = event.getLastGeneratedFormula();
        // delete ...
        // add CalcHeadMenuItem for all CalcHeads
        ICalcIterator it = (ICalcIterator) lastUnchanged.clone();
        try {
            while (it.moveDown() && it.compareTo(lastGenerated) != 0) {
                if (it.onCalcHead()) {
                    ICalcIterator iterator = (ICalcIterator) it.clone();
                    ui_control_listener_.addUIElement(new UIActionOnIterator(
                            UI_SOLVE_SHOW_CALCHEAD,
                            IUIAction.UI_CONTEXT_ONEELEMENT, iterator));
                }
            }
            if (it != null && it.onCalcHead()) {
                ICalcIterator iterator = (ICalcIterator) it.clone();
                ui_control_listener_.addUIElement(new UIActionOnIterator(
                        UI_SOLVE_SHOW_CALCHEAD,
                        IUIAction.UI_CONTEXT_ONEELEMENT, iterator));
            }
            if ((lastUnchanged.compareTo(lastGenerated) == 0)
                    && lastGenerated.onCalcHead()) {
                ICalcIterator iterator = (ICalcIterator) lastGenerated.clone();
                ui_control_listener_.addUIElement(new UIActionOnIterator(
                        UI_SOLVE_SHOW_CALCHEAD,
                        IUIAction.UI_CONTEXT_ONEELEMENT, iterator));
            }
        } catch (RemoteException e) {
            // TODO Auto-generated catch block
            e.printStackTrace();
        }

    }

    /**
     * @see isac.wsdialog.IWorksheetDialog#iterator(CalcIterator,
     *      IWorksheetDialog)
     * @see isac.interfaces.IToCalc#iterator()
     */
    public ICalcIterator iterator(CalcIterator it, IWorksheetDialog dg) {
        // WN050512...was before:
        // public ICalcIterator iterator() {
        // try {
        // return new DialogIterator((CalcIterator) calc_tree_.iterator(),
        // this);
        // } catch (Exception e) {
        // e.printStackTrace();
        // }
        // return null;
        try {
            return new DialogIterator(it, dg);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * @see isac.wsdialog.IToCalc#addListener(isac.wsdialog.IToUser)
     */
    public boolean addDataChangeListener(IToUser listener)
            throws RemoteException {
        if (datachange_listeners_.contains(listener)) {
            return false;
        } else {
            datachange_listeners_.add(listener);
        }
        return true;
    }

    public boolean registerUIControlListener(IToGUI listener)
            throws RemoteException {
        // WN0504 ???
        // if (ui_control_listener_ != null) {
        // ui_control_listener_.doUIAction(new UserAction(UI_DO_DETACH));
        // }
        ui_control_listener_ = listener;
        return true;
    }

    /**
     * Notify the Dialog Guide about interaction from the User
     * 
     * @param action
     * @return true, if the action has been processed, false if the processing
     *         the action has been denied
     * @throws RemoteException
     *  
     */
    public boolean notifyUserAction(IUserAction action)
            throws DialogProtocolException, RemoteException {
        System.out.println("in notifyUserAction");
        int request = action.getActionID();
        CalcHead calc_head = new CalcHead();

        logger_.fatal(" WS->DG: notifyUserAction request= " + request);

        switch (phase_) {
        case DIALOGPHASE_IDLE:
            throw new DialogProtocolException(request, phase_);
        case DIALOGPHASE_SPECIFY:
            if (request < UI_DUMMY_FIRST_SPECIFY
                    || request > UI_DUMMY_LAST_SPECIFY)
                throw new DialogProtocolException(request, phase_);
            break;
        case DIALOGPHASE_SOLVE:
            if (request < UI_DUMMY_FIRST_SOLVE || request > UI_DUMMY_LAST_SOLVE)
                throw new DialogProtocolException(request, phase_);
            break;
        }

        // *********** handle IUserAction's **********************

        switch (request) {
        case UI_SOLVE_SHOW_CALCHEAD:
            // in solve-phase request open a CalcHead by context-menu
            calc_head = (CalcHead) ((UserActionOnCalcElement) action)
                    .getCalcElement();
            startSpecifying(calc_head);
            break;
        // WN0508 done by ProblemBrowser; UserActions handled there
        // case UI_SPECIFY_TRY_MATCH:
        // throw new DialogNotImplementedException(request, phase_);
        // // try {
        // // try_match_(calc_head, ((UserActionOnCalcHeadCompoundID) action)
        // // .getObjectID());
        // // } catch (NotInSpecificationPhaseException e) {
        // // throw new DialogMathException(request, phase_, e);
        // // }
        // // break;
        //
        // case UI_SPECIFY_TRY_REFINE:
        // throw new DialogNotImplementedException(request, phase_);
        // // try {
        // // try_refine_(calc_head, ((UserActionOnCalcHeadCompoundID) action)
        // // .getObjectID());
        // // } catch (NotInSpecificationPhaseException e) {
        // // throw new DialogMathException(request, phase_, e);
        // // }
        // // break;
        //
        // WN0508 unused
        case UI_SPECIFY_CHANGE_VIEW:
            calc_head.setViewStyle(((UserActionOnInt) action).getInt());
            break;

        case UI_SPECIFY_FOCUS_LOST:
            System.out.println("focus lost in WSD");
            break;

        case UI_SPECIFY_COMPLETE_CALCHEAD:
            if (!calc_head.isComplete()) {
                calc_head = calc_tree_.completeCalcHead();
                doUIAction(new UIActionOnCalcElement(
                        UI_SPECIFY_COMPLETE_CALCHEAD,
                        IUIAction.UI_CONTEXT_CALCHEAD, calc_head));
            }
            break;
        case UI_SPECIFY_METHOD:
        case UI_SPECIFY_THEORY:
        case UI_SPECIFY_PROBLEM:
        case UI_SPECIFY_COMPLETE_GIVEN:
        case UI_SPECIFY_COMPLETE_FIND:
        case UI_SPECIFY_COMPLETE_RELATE:
            throw new DialogNotImplementedException(request, phase_);

        case UI_SPECIFY_CHECK_CALCHEAD:
            CalcHead ch_to_check = (CalcHead) ((UserActionOnCalcElement) action)
                    .getCalcElement();
            calc_head = calc_tree_.checkCalcHead(ch_to_check);
            doUIAction(new UIActionOnCalcElement(UI_SPECIFY_CHECK_CALCHEAD,
                    IUIAction.UI_CONTEXT_CALCHEAD, calc_head));
            break;

        case UI_SPECIFY_COMPLETE_STEPWISE:
            if (!calc_head.isComplete()) {
                calc_head = calc_tree_.nextSpecify();
                doUIAction(new UIActionOnCalcElement(
                        UI_SPECIFY_COMPLETE_STEPWISE,
                        IUIAction.UI_CONTEXT_CALCHEAD, calc_head));
            }
            break;

        case UI_SPECIFY_CLOSE_CALCHEAD:
            // this action is enabled _ONLY_ if the Panel was opened read-only
            // on a complete CalcHead just for inspection
            doUIAction(new UIAction(UI_SPECIFY_CLOSE_CALCHEAD,
                    IUIAction.UI_CONTEXT_SWITCH_PHASES));
            phase_ = DIALOGPHASE_SOLVE;// MK050812 should that have been still
            break;

        case UI_SPECIFY_RESET:
            // TOODO: not implemented in SML yet
            calc_tree_.resetCalcHead();

            // dummy CalcHead
            calc_head = new CalcHead();
            doUIAction(new UIActionOnCalcElement(UI_SPECIFY_RESET,
                    IUIAction.UI_CONTEXT_CALCHEAD, calc_head));
            break;

        case UI_SPECIFY_TO_SOLVE:
            // TOODO MK solve here
            System.out.println("Calchead start solving pressed");
        // calc_head = new CalcHead();

        // doUIAction(new UIActionOnCalcElement(UI_SPECIFY_TO_SOLVE,
        // IUIAction.UI_CONTEXT_CALCHEAD, calc_head));
        // break;
        case UI_SPECIFY_CALCULATE_ALL:
            if (!calc_head.isComplete()) {
                calc_head = calc_tree_.completeCalcHead();
                if (!calc_head.isComplete()) {
                    calc_tree_.resetCalcHead();
                    calc_head = calc_tree_.completeCalcHead();
                }
            }

            doUIAction(new UIActionOnCalcElement(UI_SPECIFY_CALCULATE_ALL,
                    IUIAction.UI_CONTEXT_CALCHEAD, calc_head));

            // if (true)
            // calc_tree_.completeCalcHead();
            // else
            // calc_tree_.checkCalcHead(calc_head);
            //
            // if (!calc_head.isComplete()) {
            // return false;
            // }

            doUIAction(new UIAction(UI_SPECIFY_CLOSE_CALCHEAD,
                    IUIAction.UI_CONTEXT_SWITCH_PHASES));
            startSolving();

        // --------------solving phase------------------------------------
        case UI_SOLVE_CALCULATE_1:
            // handle the CalcHead such that its HeadLine can be
            // presented like a CalcFormula
            if (active_formula_.onCalcHead()) {
                calc_tree_.completeCalcHead();
                try {
                    calc_tree_.startSolving();
                } catch (AutoCalculateException e1) {
                    e1.printStackTrace();
                }
            } else {
                calc_tree_.autoCalculate(IToCalc.SCOPE_CALCULATION, 1);
            }
            break;
        case UI_SOLVE_CALCULATE_ALL:
            switch (request) {
            case UI_SPECIFY_CALCULATE_ALL:
            case UI_SOLVE_CALCULATE_ALL:
                calc_tree_.autoCalculate(IToCalc.SCOPE_CALCULATION, 0);
                break;
            case UI_SPECIFY_TO_SOLVE:
            default:
            }
            break;

        case UI_SOLVE_CALCULATE_SUBPROBLEM:
            calc_tree_.autoCalculate(IToCalc.SCOPE_SUBPROBLEM, 0);
            break;

        case UI_SOLVE_EDIT_ACTIVE_FORMULA:
            ui_control_listener_.doUIAction(new UserAction(UI_DO_EDIT_FORMULA));
            break;

        case UI_SOLVE_EDIT_ACTIVE_FORMULA_COMPLETE:
            System.out.println("**********XXX****************");
            CalcFormula formula = (CalcFormula) ((UserActionOnCalcElement) action)
                    .getCalcElement();
            // if the position is null, the formula did not exist, so we try to
            // append a new formula
            if (formula.getPosition() == null) {
                calc_tree_.appendFormula(formula);
            } else {
                active_formula_ = new CalcIterator(calc_tree_, formula
                        .getPosition());
                calc_tree_.replaceFormula(formula);
            }
            break;

        case UI_SOLVE_APPEND_USER_FORMULA:
            // appendFormula((Formula) ((UserActionOnCalcElement)
            // action).getCalcElement());
            ui_control_listener_
                    .doUIAction(new UserAction(UI_DO_APPEND_FORMULA));
            break;

        case UI_SOLVE_MOVE_ACTIVE_FORMULA:
            calc_tree_.moveActiveFormula(((UserActionOnIterator) action)
                    .getPosition());
            break;

        case UI_SOLVE_GET_PROPOSED_TACTIC:
            throw new DialogNotImplementedException(request, phase_);

        case UI_SOLVE_GET_APPLICABLE_TACTICS:
            // throw new DialogNotImplementedException(request, phase_);
            System.out.println("!!!!!!!!!get applicable tactics!!!!!!");
        case UI_SOLVE_SET_NEXT_TACTIC:
            calc_tree_
                    .setNextTactic((Tactic) ((UserActionOnCalcElement) action)
                            .getCalcElement());
            // WN041126 extracted from isac.bridge.CalcTree#setNextTactic #@#
            // autoCalculate(SCOPE_CALCULATION, 1);
            break;

        case UI_SOLVE_HELP_ENTERING_FORMULA:
            throw new DialogNotImplementedException(request, phase_);

        case UI_SOLVE_SHOW_ASSUMPTIONS:
            throw new DialogNotImplementedException(request, phase_);

        case UI_SOLVE_SHOW_INTERMEDIATE_STEPS:
            throw new DialogNotImplementedException(request, phase_);

        case UI_SPECIFY_OPEN_CALCHEAD:
            System.out.println("Pressed!");
            break;

        default:
            throw new DialogUnknownActionException(request, phase_);
        }
        return true;
    }

    private void rmiBind() {
        if (System.getSecurityManager() == null) {
            System.setSecurityManager(new RMISecurityManager());
        }

        try {
            LocateRegistry.createRegistry(1099);
        } catch (java.rmi.RemoteException exc2) {
            System.err.println("can not create registry: " + exc2.getMessage());
        }

        String name = "//localhost/isac-WorksheetDialog";
        try {
            System.out.println("try to bind as " + name);
            Naming.rebind(name, this);
            System.out.println("Dialog Guide bound to " + name);
            // WN040906 was Object Manager ...
        } catch (java.rmi.ConnectException e) {
            System.err.println("failed to contact as " + name
                    + " (creating RMI-Server on localhost: 1099)");
        } catch (RemoteException e) {
            e.printStackTrace();
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }
    }

    /**
     * @see isac.wsdialog.IWorksheetDialog#notifyGUIReady()
     */
    public void notifyGUIReady() throws RemoteException {
        worksheet_ready_ = true;
    }

    /**
     * @see isac.interfaces.IToGUI#doUIAction(isac.wsdialog.IUIAction)
     */
    public synchronized void doUIAction(final IUIAction action)
            throws RemoteException {
        Thread t = new Thread(new Runnable() {

            public void run() {
                try {
                    while (worksheet_ready_ == false) {
                        try {
                            Thread.sleep(1000);
                        } catch (InterruptedException e1) {
                            // TODO Auto-generated catch block
                            e1.printStackTrace();
                        }
                    }
                    ui_control_listener_.doUIAction(action);
                    // ???worksheet_ready_ = false;
                } catch (RemoteException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        });
        t.run();
    }

    /**
     * request the gui to add an interactive element
     * 
     * @see isac.interfaces.IToGUI#addUIElement(isac.wsdialog.IUIAction)
     */
    public synchronized void addUIElement(final IUIAction action)
            throws RemoteException {
        Thread t = new Thread(new Runnable() {

            public void run() {
                try {
                    while (worksheet_ready_ == false) {
                        try {
                            Thread.sleep(1000);
                        } catch (InterruptedException e1) {
                            e1.printStackTrace();
                        }
                    }
                    ui_control_listener_.addUIElement(action);
                } catch (RemoteException e) {
                    e.printStackTrace();
                }
            }
        });
        t.run();
    }

    /**
     * request the gui to remove an interactive element
     * 
     * @see isac.interfaces.IToGUI#removeUIElement(isac.wsdialog.IUIAction)
     */
    public synchronized void removeUIElement(final IUIAction action)
            throws RemoteException {
        Thread t = new Thread(new Runnable() {

            public void run() {
                try {
                    while (worksheet_ready_ == false) {
                        try {
                            Thread.sleep(1000);
                        } catch (InterruptedException e1) {
                            // TODO Auto-generated catch block
                            e1.printStackTrace();
                        }
                    }
                    ui_control_listener_.removeUIElement(action);
                    // ???worksheet_ready_ = false;
                } catch (RemoteException e) {
                    // TODO Auto-generated catch block
                    e.printStackTrace();
                }
            }
        });
        t.run();
    }

    /**
     * @see isac.wsdialog.IBrowserDialogToWorksheetDialog#tryMatchProblem(isac.util.formulae.KEStoreID)
     */
    public Match tryMatchProblem(ProblemID id) {
        try {
            if (active_formula_.onCalcHead())
                return active_formula_.tryMatchProblem(id);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * @see isac.wsdialog.IBrowserDialogToWorksheetDialog#tryMatchMethod(isac.util.formulae.KEStoreID)
     */
    public Match tryMatchMethod(MethodID id) {
        try {
            if (active_formula_.onCalcHead())
                return active_formula_.tryMatchMethod(id);
        } catch (RemoteException e) {
            e.printStackTrace();
        }
        return null;
    }

}